#include "TEMPRAsys.h"

#define	SECTOR	512
#define TEMPRA_DEVICE_NAME			L"\\Device\\TEMPRA"
#define TAG_MEM                     '1maR'  // "Ram1"
#define TAG_SCR                     '2maR'  // "Ram2"

#if DBG
VOID DbgStart();
VOID DbgText(IN PCHAR);
VOID DbgWrite(IN ULONG,IN PBOOT_SECTOR);
#endif

typedef struct _DEVICE_EXTENSION {
	ULONG						DiskInfo;		// fBXN
	PUCHAR						DiskImage;		// TEMPRAp̃nh
	ULONG						DiskSize;		// TEMPRAp̃oCg
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;


//-------------------------------------------------
NTSTATUS FormatTempraDrive (
	IN PDEVICE_OBJECT	devObj
	,IN ULONG			DiskInfo
	)
{
	NTSTATUS			status;
	PDRIVER_OBJECT	DriverObject;
	PDEVICE_EXTENSION	devExt;

	ULONG			uMemSize;
	ULONG			nAllocUnit;
	ULONG			nRootDir;
    ULONG			fatType;
    ULONG			fatEntries;
    ULONG			fatSectorCnt;

	PBOOT_SECTOR bootSector;
	PBOOT_SECTOR_32 bootSector32;
    PUCHAR firstFatSector;
    PUCHAR pText,fsinfoSector;
    PDIR_ENTRY rootDir;
	int nn;
	DISK_INFO *disk_info;

	KdPrint(("FormatTempraDrive - IN\n"));

	status=STATUS_SUCCESS;
	DriverObject=(PDRIVER_OBJECT)devObj->DriverObject;
	devExt = (PDEVICE_EXTENSION)devObj->DeviceExtension;

	disk_info = (PVOID)&DiskInfo;
	KdPrint(("DiskInfo=0x%08x\n",DiskInfo));
	KdPrint(("disk_info->sizeMB=%d\n",disk_info->sizeMB));
	uMemSize = (ULONG)(disk_info->sizeMB)*(1024*1024);
	if((disk_info->option & 0x08) == 0) fatType=16; else fatType=32;
	if(fatType==32){
		nAllocUnit=1;
		nn=(ULONG)(disk_info->option & 0x07);
		while(nn>0){
			if( ((uMemSize/512 - 2)/(2*nAllocUnit)) /*+ 2*/ < 65536 ) break;// ׽<=0xffff̓_
			nAllocUnit *= 2;
			nn--;
		}
	}
	else{
		nAllocUnit=2;
		while(uMemSize > /*65536*/(ULONG)0xfff6*SECTOR*nAllocUnit) nAllocUnit *= 2;
	}

	/*@fBNgGgZbg@*/
	if(fatType==32){
		nRootDir = 0;
	}
	else{
		nRootDir = 64*(ULONG)(disk_info->option & 0x07) + 128;
		nn = SECTOR/sizeof(DIR_ENTRY);	// DIR_ENTRY32޲ĂȂ̂łPZN^(512޲)16
		if (nRootDir & (nn - 1)) {	// 16̔{łȂ
			// 16̔{ɂȂ悤؂グ
			nRootDir = (nRootDir + (nn - 1)) &  (~(nn - 1));
		}
	}

	/*@foCX@*/
	devObj->Flags |= DO_DIRECT_IO;
	devObj->AlignmentRequirement = FILE_WORD_ALIGNMENT;

	/*@̊mۂƏ@*/
	devExt->DiskSize = uMemSize;
	if(uMemSize==0){
		devExt->DiskImage = NULL;
	}
	else{
		if((disk_info->option & 0x10) == 0){
			KdPrint(("NonPagedPool  uMemSize=%d bytes\n", uMemSize));
			devExt->DiskImage = ExAllocatePoolWithTag(NonPagedPool, uMemSize, TAG_MEM);
		}
		else{
			KdPrint(("PagedPool  uMemSize=%d bytes\n", uMemSize));
			devExt->DiskImage = ExAllocatePoolWithTag(PagedPool, uMemSize, TAG_MEM);
		}
	}
	if (devExt->DiskImage == NULL) {	/*@mێs@*/
		KdPrint(("(ERROR) ExAllocatePool\n"));
		devExt->DiskSize = 0;
		status = STATUS_INSUFFICIENT_RESOURCES;
		return status;
	}
	RtlZeroBytes(devExt->DiskImage, uMemSize);


	/*@u[gZN^Zbg@*/
	bootSector=(PBOOT_SECTOR)devExt->DiskImage;
	bootSector32=(PBOOT_SECTOR_32)devExt->DiskImage;

    // We need to have the 0xeb and 0x90 since this is one of the
    // checks the file system recognizer uses
    bootSector->bsJump[0] = 0xeb;
    bootSector->bsJump[1] = 0x58;
    bootSector->bsJump[2] = 0x90;

    // Set OemName to "TEMPRA  "
    bootSector->bsOemName[0] = 'T';
    bootSector->bsOemName[1] = 'E';
    bootSector->bsOemName[2] = 'M';
    bootSector->bsOemName[3] = 'P';
    bootSector->bsOemName[4] = 'R';
    bootSector->bsOemName[5] = 'A';
    bootSector->bsOemName[6] = ' ';
    bootSector->bsOemName[7] = ' ';

    bootSector->bsBytesPerSec = SECTOR;
    bootSector->bsSecPerClus  = (UCHAR)nAllocUnit;
    bootSector->bsResSectors  = (fatType==32) ? 2 : 1;	/*\(FAT32ްľFSINFO)*/
    bootSector->bsFATs        = 1;						/*FAT*/

	if(fatType!=32){
	    bootSector->bsRootDirEnts   = (USHORT)nRootDir;		/*ٰިڸشؐ*/
	    bootSector->bsSectors       = 0;
		if(uMemSize/512 < 0x10000){	/* 32MB */
			bootSector->bsSectors   = (USHORT)(uMemSize/SECTOR);/*S*/
		}

	    bootSector->bsMedia         = 0xf8;					/*ިID*/

	    // Calculate number of sectors required for FAT
	    //   (( S - \ - ٰިڸشؐ/16 ) / ׽辸) + 2
	    nn = uMemSize/512 - bootSector->bsResSectors - bootSector->bsRootDirEnts/16;
	    fatEntries = (nn / bootSector->bsSecPerClus) + 2;

	    // Choose between 12 and 16 bit FAT based on number of clusters we need to map
	    if (fatEntries > 4087) {
	        fatSectorCnt = (fatEntries * 2 + 511) / 512;
	        fatEntries   = fatEntries + fatSectorCnt;
	        fatSectorCnt = (fatEntries * 2 + 511) / 512;
	    }
	    else {
	        fatType =  12;
	        fatSectorCnt = (((fatEntries * 3 + 1) / 2) + 511) / 512;
	        fatEntries   = fatEntries + fatSectorCnt;
	        fatSectorCnt = (((fatEntries * 3 + 1) / 2) + 511) / 512;
	    }

	    bootSector->bsFATsecs            = (USHORT)fatSectorCnt;	/*FAT*/
	}
	else{	// FAT32
	    bootSector->bsRootDirEnts   = 0;		/*ٰިڸشؐ(FAT32͏ɂO)*/
	    bootSector->bsSectors       = 0;		/*S(FAT32͏ɂO)*/
	    bootSector->bsMedia         = 0xf8;		/*ިID*/
	    bootSector->bsFATsecs       = 0;		/*FAT(FAT32͏ɂO)*/
	}

    bootSector->bsSecPerTrack        = 0;			// not use
    bootSector->bsHeads              = 0;			// not use
    bootSector->bsHiddenSecs         = 0;			/*Bꂽ*/
    bootSector->bsHugeSectors        = 0;
	if(uMemSize/512 >= 0x10000){	/* 32MBȏ */
		bootSector->bsHugeSectors    = uMemSize/SECTOR;	/*S*/
	}

	if(fatType!=32){	// FAT12/16
	    bootSector->bsBootSignature      = 0x29;
	    bootSector->bsVolumeID           = 0x12345678;
	}
	else{	// FAT32
	    // Calculate number of sectors required for FAT32
	    //   (( S - \ ) / ׽辸) + 2
	    nn = uMemSize/512 - bootSector->bsResSectors;
	    fatEntries = (nn / bootSector->bsSecPerClus) + 2;
		// FAT3232bit(4byte)Ȃ̂1(512byte)128
	    fatSectorCnt = (fatEntries- 1) / 128 + 1;

	    bootSector32->bsFATsecs32        = fatSectorCnt;
	    bootSector32->bsExtFlags32       = 0;
	    bootSector32->bsFilesysVer32     = 0;
	    bootSector32->bsRootClus32       = 2;
	    bootSector32->bsFsInfo32         = 1;
	    bootSector32->bsBootBack32       = 0xffff;
	    bootSector32->bsBootSignature    = 0x29;
	    bootSector32->bsVolumeID         = 0x12345678;
	}

    // Set Label to "TEMPRA-DISK"
	if(fatType!=32) pText=bootSector->bsLabel; else pText=bootSector32->bsLabel;
    pText[0]  = 'T';
    pText[1]  = 'E';
    pText[2]  = 'M';
    pText[3]  = 'P';
    pText[4]  = 'R';
    pText[5]  = 'A';
    pText[6]  = '-';
    pText[7]  = 'D';
    pText[8]  = 'I';
    pText[9]  = 'S';
    pText[10] = 'K';

	if(fatType!=32) pText=bootSector->bsFileSystemType; else pText=bootSector32->bsFileSystemType;
    // Set FileSystemType to "FAT1?   "
    pText[0] = 'F';
    pText[1] = 'A';
    pText[2] = 'T';
    pText[3] = ( fatType == 32 ) ? '3' : '1';
    pText[4] = ( fatType == 16 ) ? '6' : '2';
    pText[5] = ' ';
    pText[6] = ' ';
    pText[7] = ' ';

    bootSector->bsSig2[0] = 0x55;
    bootSector->bsSig2[1] = 0xAA;

	if(fatType==32){	// FSINFO(FAT32)
    	fsinfoSector      = (PUCHAR)(bootSector + 1);
	    fsinfoSector[0x000] = 'R';
	    fsinfoSector[0x001] = 'R';
	    fsinfoSector[0x002] = 'a';
	    fsinfoSector[0x003] = 'A';
	    fsinfoSector[0x1e4] = 'r';
	    fsinfoSector[0x1e5] = 'r';
	    fsinfoSector[0x1e6] = 'A';
	    fsinfoSector[0x1e7] = 'a';
	    fsinfoSector[0x1e8] = 0xff;
	    fsinfoSector[0x1e9] = 0xff;
	    fsinfoSector[0x1ea] = 0xff;
	    fsinfoSector[0x1eb] = 0xff;
	    fsinfoSector[0x1ec] = 0x02;
	    fsinfoSector[0x1ed] = 0x00;
	    fsinfoSector[0x1ee] = 0x00;
	    fsinfoSector[0x1ef] = 0x00;
	    fsinfoSector[0x1fe] = 0x55;
	    fsinfoSector[0x1ff] = 0xaa;
	}

    // The FAT is located immediately following the boot sector.
    firstFatSector    = (PUCHAR)(bootSector + bootSector->bsResSectors);
    firstFatSector[0] = 0xf8;
    firstFatSector[1] = 0xff;
    firstFatSector[2] = 0xff;

    if (fatType != 12) firstFatSector[3] = 0xff;
    if (fatType == 32){
		firstFatSector[4]  = 0xff;
		firstFatSector[5]  = 0xff;
		firstFatSector[6]  = 0xff;
		firstFatSector[7]  = 0xff;
		firstFatSector[8]  = 0xff;
		firstFatSector[9]  = 0xff;
		firstFatSector[10] = 0xff;
		firstFatSector[11] = 0x0f;
	}

    // The Root Directory follows the FAT
	//@{[x
    rootDir = (PDIR_ENTRY)(bootSector + bootSector->bsResSectors + fatSectorCnt);
    rootDir->deName[0]      = 'T';
    rootDir->deName[1]      = 'E';
    rootDir->deName[2]      = 'M';
    rootDir->deName[3]      = 'P';
    rootDir->deName[4]      = 'R';
    rootDir->deName[5]      = 'A';
    rootDir->deName[6]      = '-';
    rootDir->deName[7]      = 'D';
    rootDir->deExtension[0] = 'I';
    rootDir->deExtension[1] = 'S';
    rootDir->deExtension[2] = 'K';
    rootDir->deAttributes   = 0x08;

	KdPrint(("FormatTempraDrive - OUT\n"));
	return status;
}


//-------------------------------------------------
NTSTATUS TempraCreateClose(
	IN PDEVICE_OBJECT	devObj
	,IN PIRP			pIrp
	)
{
	KdPrint(("TempraCreateClose --- IN\n"));
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	KdPrint(("TempraCreateClose --- OUT\n"));
	return STATUS_SUCCESS;
}


//-------------------------------------------------
NTSTATUS TempraReadWrite(
	IN PDEVICE_OBJECT	devObj
	,IN PIRP			pIrp
	)
{
	PDEVICE_EXTENSION	devExt;
	PIO_STACK_LOCATION	irpStack;
	PUCHAR				currentAddress;
    ULONG               information;

	KdPrint(("TempraReadWrite --- IN\n"));
	/* \̐擪|C^擾 */
	devExt = devObj->DeviceExtension;
	/* X^bNւ̃|C^擾 */
	irpStack = IoGetCurrentIrpStackLocation(pIrp);

	if(devExt->DiskSize==0){
		KdPrint(("  No media in device\n"));
		pIrp->IoStatus.Status = STATUS_NO_MEDIA_IN_DEVICE;
		pIrp->IoStatus.Information = 0;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);
		return STATUS_NO_MEDIA_IN_DEVICE;
	}

    if(RtlLargeIntegerGreaterThan(
            RtlLargeIntegerAdd( 
				irpStack->Parameters.Read.ByteOffset,
				RtlConvertUlongToLargeInteger(irpStack->Parameters.Read.Length)),
            RtlConvertUlongToLargeInteger(devExt->DiskSize)) ||		/* fBXNeʂ𒴂 */
        (irpStack->Parameters.Read.Length & (SECTOR-1)) ||			/* ZN^TCY̔{łȂ */
		(irpStack->Parameters.Read.ByteOffset.LowPart & (SECTOR-1))	/* JnʒuZN^TCY̔{łȂ */
	){
		KdPrint(("(ERROR) Invalid I/O parameter\n"));
		pIrp->IoStatus.Status = STATUS_INVALID_PARAMETER;
		pIrp->IoStatus.Information = 0;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);
//		IoReleaseRemoveLock(&devExt->RemoveLock, pIrp);
		return STATUS_INVALID_PARAMETER;
	}

	if(pIrp->MdlAddress != NULL){
	    currentAddress = MmGetSystemAddressForMdlSafe( pIrp->MdlAddress, NormalPagePriority );
		if(currentAddress == NULL){
			KdPrint(("(ERROR) MmGetSystemAddressForMdlSafe\n"));
			pIrp->IoStatus.Status = STATUS_INSUFFICIENT_RESOURCES;
			pIrp->IoStatus.Information = 0;
			IoCompleteRequest(pIrp, IO_NO_INCREMENT);
//			IoReleaseRemoveLock(&devExt->RemoveLock, pIrp);
			return STATUS_INSUFFICIENT_RESOURCES;
		}
	}

    information = irpStack->Parameters.Read.Length;
	switch (irpStack->MajorFunction)
	{
	case IRP_MJ_READ:		/* READ*/
		KdPrint(("  (MJ_READ)\n"));
        RtlMoveMemory(
            currentAddress,		// Dst
			devExt->DiskImage + irpStack->Parameters.Read.ByteOffset.LowPart,	// Src
            information			// Length
		);
		break;

	case IRP_MJ_WRITE:		/* WRITE */
		KdPrint(("  (MJ_WRITE)\n"));
        RtlMoveMemory(
			devExt->DiskImage + irpStack->Parameters.Read.ByteOffset.LowPart,	// Dst
            currentAddress,		// Src
            information			// Length
		);
		break;

	default:
		KdPrint(("  (IRP 0x%02x)\n", (UCHAR)(irpStack->MajorFunction)));
		information = 0;
		break;
	}

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = information;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	KdPrint(("TempraReadWrite --- OUT\n"));
	return STATUS_SUCCESS;
}


//-------------------------------------------------
NTSTATUS TempraDeviceControl (
	IN PDEVICE_OBJECT	devObj
	,IN PIRP 			pIrp
	)
{
	PDEVICE_EXTENSION	devExt;
	PIO_STACK_LOCATION	irpStack;
	NTSTATUS			status;
	PBOOT_SECTOR bootSector;
	ULONG ctrlCode;

	KdPrint(("TempraDeviceControl --- IN\n"));

	devExt = (PDEVICE_EXTENSION)devObj->DeviceExtension;
	irpStack = IoGetCurrentIrpStackLocation(pIrp);
	ctrlCode=(ULONG)(irpStack->Parameters.DeviceIoControl.IoControlCode);

	pIrp->IoStatus.Information = 0;
	status = STATUS_SUCCESS;

	switch (ctrlCode) {
	case IOCTL_DISK_GET_DRIVE_GEOMETRY:
	case IOCTL_DISK_GET_MEDIA_TYPES:
	case IOCTL_STORAGE_GET_MEDIA_TYPES:
	{	/* fBXNWIg */
//		ULONG nn;
		PDISK_GEOMETRY pBuf;

		if(ctrlCode==IOCTL_DISK_GET_DRIVE_GEOMETRY){
			KdPrint(("  (GEOMETRY %08x)\n",ctrlCode));
		}
		else{
			KdPrint(("  (MEDIA_TYPES %08x)\n",ctrlCode));
		}
		if (irpStack->Parameters.DeviceIoControl.OutputBufferLength < sizeof(DISK_GEOMETRY)) {		/*@TCYs@*/
			status = STATUS_BUFFER_TOO_SMALL;
			KdPrint(("(ERROR)TempraDeviceControlGeometry:size too small\n"));
			break;
		}

		if(devExt->DiskSize==0){
			status = STATUS_INVALID_DEVICE_REQUEST;
		}
		else{
			bootSector=(PBOOT_SECTOR)devExt->DiskImage;
			pBuf = (PDISK_GEOMETRY)(pIrp->AssociatedIrp.SystemBuffer);
				RtlZeroMemory( pBuf, sizeof(DISK_GEOMETRY) );
			pBuf->MediaType = bootSector->bsMedia;		/* fBA^Cv */
			pBuf->BytesPerSector    = bootSector->bsBytesPerSec;
			pIrp->IoStatus.Information = sizeof(DISK_GEOMETRY);
		}
		break;
	}

	case IOCTL_DISK_GET_PARTITION_INFO:
	{	/* p[e[V */
		PPARTITION_INFORMATION pBuf;

		KdPrint(("  (PARTITION %08x)\n",ctrlCode));
		if (irpStack->Parameters.DeviceIoControl.OutputBufferLength < sizeof(PARTITION_INFORMATION)) {
			/* TCYs */
			status = STATUS_INVALID_PARAMETER;
			KdPrint(("(ERROR)TempraDeviceControlGetPartInfo:size too small\n"));
			break;
		}
		if(devExt->DiskSize==0){
			status = STATUS_INVALID_DEVICE_REQUEST;
		}
		else{
			bootSector=(PBOOT_SECTOR)devExt->DiskImage;
			pBuf = (PPARTITION_INFORMATION)(pIrp->AssociatedIrp.SystemBuffer);
			pBuf->PartitionType =
				(bootSector->bsFileSystemType[4]=='2') ? PARTITION_FAT_12 : PARTITION_FAT_16;
			pBuf->BootIndicator       = FALSE;		/* u[gs */
			pBuf->RecognizedPartition = TRUE;		/* p[e[VF */
			pBuf->RewritePartition    = FALSE;		/* ďݕsp[e[V */
			pBuf->StartingOffset      = RtlConvertUlongToLargeInteger(0);	/* p[e[VJnʒu */
			pBuf->PartitionLength     = RtlConvertUlongToLargeInteger(devExt->DiskSize);	/*  */
			pBuf->HiddenSectors       = bootSector->bsHiddenSecs;	/* BZN^ */
			pBuf->PartitionNumber     = 1;			/* p[e[V */
			pIrp->IoStatus.Information = sizeof(PARTITION_INFORMATION);
		}
		break;
	}

	case IOCTL_DISK_CHECK_VERIFY:
	case IOCTL_STORAGE_CHECK_VERIFY:
	case IOCTL_STORAGE_CHECK_VERIFY2:
	// ̏͂Ȃƃ_iTEMPRA-DISKłȂ۰ިƕ\걸łȂj
		KdPrint(("  (CHECK_VERIFY %08x)\n",ctrlCode));
		break;

	case IOCTL_DEVICE_OPEN:
	{
		PULONG AllocSize;

		KdPrint((" yIOCTL_DEVICE_OPEN %08xz\n",ctrlCode));
#if DBG
		DbgText("(IOCTL_DEVICE_OPEN)");
#endif

		AllocSize = (PULONG)pIrp->AssociatedIrp.SystemBuffer;
		*AllocSize=devExt->DiskSize;

		pIrp->IoStatus.Information = sizeof(ULONG);
		pIrp->IoStatus.Status = STATUS_SUCCESS;
//		PsRevertToSelf();
		break;
	}

	default:
		//	Unknown IOCTL request
		KdPrint(("  (Unknown %08x)\n",ctrlCode));
		status = STATUS_INVALID_DEVICE_REQUEST;
	}

	pIrp->IoStatus.Status = status;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);

	KdPrint(("TempraDeviceControl --- OUT\n"));
	return status;
}

//-------------------------------------------------
VOID TempraUnloadDriver (
	IN PDRIVER_OBJECT DriverObject
	)
{
	UNICODE_STRING		unicodeBuf;
	PDEVICE_OBJECT	devObj;
	PDEVICE_EXTENSION	devExt;
	PULONG DiskInfo;
	DISK_INFO *disk_info;
	WCHAR wcharBuf[]={L"\\??\\A:"};
	WCHAR alphabet[]={L"ABCDEFGHIJKLMNOPQRSTUVWXYZ"};

	KdPrint(("TempraUnloadDriver - IN\n"));
#if DBG
	DbgText(("TempraUnloadDriver - IN"));
#endif

	devObj=DriverObject->DeviceObject;
	devExt = (PDEVICE_EXTENSION)devObj->DeviceExtension;

	if (devExt->DiskSize!=0) {
		ExFreePool(devExt->DiskImage);
		devExt->DiskSize=0;
		devExt->DiskImage = NULL;
	}

	//	Delete Default Symbolic Link
	DiskInfo = (PULONG)&(devExt->DiskInfo);
	KdPrint(("  DiskInfo %08x\n",*DiskInfo));
	disk_info = (PVOID)&(devExt->DiskInfo);

	wcharBuf[4]=alphabet[(ULONG)(disk_info->drive)-(ULONG)('A')];
	RtlInitUnicodeString(&unicodeBuf, wcharBuf);
	KdPrint(("IoDeleteSymbolicLink(%wZ)\n", &unicodeBuf));
	IoDeleteSymbolicLink(&unicodeBuf);

	//	Delete the device object
	IoDeleteDevice(devObj);

#if DBG
	DbgText(("TempraUnloadDriver - OUT"));
#endif
	KdPrint(("TempraUnloadDriver - OUT\n"));
	return;
}

//-------------------------------------------------
NTSTATUS DriverEntry (
	IN PDRIVER_OBJECT	DriverObject
	,IN PUNICODE_STRING	RegistryPath
	)
{
	NTSTATUS			status;
	UNICODE_STRING		device_name;	/* L"\\Device\\TEMPRA" */
	UNICODE_STRING		unicodeBuf;		/* L"\\DosDevices\\X:" */
	PDEVICE_OBJECT		devObj;
	PDEVICE_EXTENSION	devExt;

	WCHAR dos_device[]={L"\\DosDevices\\A:"};
	WCHAR alphabet[]={L"ABCDEFGHIJKLMNOPQRSTUVWXYZ"};

	ULONG DiskInfo,defDiskInfo;
	DISK_INFO *disk_info;
	RTL_QUERY_REGISTRY_TABLE    rtlQueryRegTbl[ 2 + 1 ];

	KdPrint(("_MSC_VER=%08x\n",_MSC_VER));
	KdPrint(("DriverEntry - IN\n"));
	ASSERT(DriverObject);

#if DBG
	DbgStart();
	DbgText("DriverEntry - IN");
#endif

// WXgǂݍ
	RtlZeroMemory( rtlQueryRegTbl, sizeof(rtlQueryRegTbl) );

#if DBG
	{
		WCHAR wcharBuf[10];
		unicodeBuf.Buffer=(PWSTR)&wcharBuf;
		unicodeBuf.MaximumLength=sizeof(wcharBuf);
		unicodeBuf.Length=0;
		rtlQueryRegTbl[0].Flags         = RTL_QUERY_REGISTRY_DIRECT;
		rtlQueryRegTbl[0].Name          = L"DisplayName";
		rtlQueryRegTbl[0].EntryContext  = &unicodeBuf;
		rtlQueryRegTbl[0].DefaultType   = REG_SZ;
		rtlQueryRegTbl[0].DefaultData   = L"TEMPRA";
		rtlQueryRegTbl[0].DefaultLength = 0;

		status = RtlQueryRegistryValues(
			RTL_REGISTRY_ABSOLUTE | RTL_REGISTRY_OPTIONAL,
			RegistryPath->Buffer,
			rtlQueryRegTbl,
			NULL,
			NULL
			);
		if(status!=STATUS_SUCCESS) RtlInitUnicodeString(&unicodeBuf, L"error");
		KdPrint(("DisplayName: %wZ\n", &unicodeBuf));
	}
#endif

	rtlQueryRegTbl[0].Flags         = RTL_QUERY_REGISTRY_SUBKEY;
	rtlQueryRegTbl[0].Name          = L"Parameters";
	rtlQueryRegTbl[0].EntryContext  = NULL;
	rtlQueryRegTbl[0].DefaultType   = (ULONG)NULL;
	rtlQueryRegTbl[0].DefaultData   = NULL;
	rtlQueryRegTbl[0].DefaultLength = (ULONG)NULL;

	defDiskInfo=0x00520020;
	rtlQueryRegTbl[1].Flags         = RTL_QUERY_REGISTRY_DIRECT;
	rtlQueryRegTbl[1].Name          = L"DiskInfo";
	rtlQueryRegTbl[1].EntryContext  = &DiskInfo;
	rtlQueryRegTbl[1].DefaultType   = REG_DWORD;
	rtlQueryRegTbl[1].DefaultData   = &defDiskInfo;
	rtlQueryRegTbl[1].DefaultLength = sizeof(ULONG);

	status = RtlQueryRegistryValues(
		RTL_REGISTRY_ABSOLUTE | RTL_REGISTRY_OPTIONAL,
		RegistryPath->Buffer,
		rtlQueryRegTbl,
		NULL,
		NULL
		);
	if(status!=STATUS_SUCCESS) DiskInfo=0;
	disk_info = (PVOID)&DiskInfo;

// foCX쐬
	RtlInitUnicodeString(&device_name, TEMPRA_DEVICE_NAME);

	if((disk_info->option & 0x20) == 0){
		status = IoCreateDevice(
			DriverObject,sizeof(DEVICE_EXTENSION),&device_name,
			FILE_DEVICE_DISK, 0,					//	n[hfBXN
//			FILE_DEVICE_DISK,FILE_REMOVABLE_MEDIA,	//	[ofBXN
			FALSE,&devObj);
	}
	else{
		status = IoCreateDevice(	// fBXN
			DriverObject,sizeof(DEVICE_EXTENSION),&device_name,
			FILE_DEVICE_VIRTUAL_DISK,(FILE_DEVICE_SECURE_OPEN),
			FALSE,&devObj);
	}
	if (!NT_SUCCESS(status)) {
		KdPrint(("(ERROR) IoCreateDevice() 0x%x\n", status));
		return status;
	}

// |C^擾A
	devExt = (PDEVICE_EXTENSION)devObj->DeviceExtension;
	RtlZeroMemory(devExt, sizeof(DEVICE_EXTENSION));
	devExt->DiskInfo=DiskInfo;

// RAM-DISKtH[}bg
	FormatTempraDrive(devObj,DiskInfo);
#if DBG
	DbgWrite(DiskInfo,(PBOOT_SECTOR)devExt->DiskImage);
#endif

// N쐬
	dos_device[12]=alphabet[(ULONG)(disk_info->drive)-(ULONG)('A')];
	RtlInitUnicodeString(&unicodeBuf, dos_device);

	KdPrint(("IoCreateSymbolicLink(%wZ,%wZ)\n", &unicodeBuf,&device_name));
	status = IoCreateSymbolicLink(&unicodeBuf, &device_name);
	if (!NT_SUCCESS(status)) {
		KdPrint(("(ERROR) IoCreateSymbolicLink() 0x%x\n", status));
		IoDeleteDevice(devObj);
		return status;
	}

//	Setup dispatch table
	DriverObject->MajorFunction[IRP_MJ_CREATE]		   = TempraCreateClose;
	DriverObject->MajorFunction[IRP_MJ_CLOSE]		   = TempraCreateClose;
	DriverObject->MajorFunction[IRP_MJ_READ]		   = TempraReadWrite;
	DriverObject->MajorFunction[IRP_MJ_WRITE]		   = TempraReadWrite;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = TempraDeviceControl;
	DriverObject->DriverUnload = TempraUnloadDriver;

#if DBG
	DbgText(("DriverEntry - OUT"));
#endif
	KdPrint(("DriverEntry - OUT\n"));

	return STATUS_SUCCESS;
}
